MODULE traadv_tvd
   !!==============================================================================
   !!                       ***  MODULE  traadv_tvd  ***
   !! Ocean active tracers:  horizontal & vertical advective trend
   !!==============================================================================
   !! History :       !  95-12  (L. Mortier)  Original code
   !!                 !  00-01  (H. Loukos)  adapted to ORCA 
   !!                 !  00-10  (MA Foujols E.Kestenare)  include file not routine
   !!                 !  00-12  (E. Kestenare M. Levy)  fix bug in trtrd indexes
   !!                 !  01-07  (E. Durand G. Madec)  adaptation to ORCA config
   !!            8.5  !  02-06  (G. Madec)  F90: Free form and module
   !!            9.0  !  04-01  (A. de Miranda, G. Madec, J.M. Molines ): advective bbl
   !!            9.0  !  08-04  (S. Cravatte) add the i-, j- & k- trends computation
   !!            " "  !  05-11  (V. Garnier) Surface pressure gradient organization
   !!----------------------------------------------------------------------


   !!----------------------------------------------------------------------
   !!   tra_adv_tvd  : update the tracer trend with the horizontal
   !!                  and vertical advection trends using a TVD scheme
   !!   nonosc       : compute monotonic tracer fluxes by a nonoscillatory
   !!                  algorithm 
   !!----------------------------------------------------------------------
   USE oce             ! ocean dynamics and active tracers
   USE dom_oce         ! ocean space and time domain
   USE trdmod          ! ocean active tracers trends 
   USE trdmod_oce      ! ocean variables trends
   USE in_out_manager  ! I/O manager
   USE dynspg_oce      ! choice/control of key cpp for surface pressure gradient
   USE trabbl          ! Advective term of BBL
   USE lib_mpp
   USE lbclnk          ! ocean lateral boundary condition (or mpp link) 
   USE diaptr          ! poleward transport diagnostics
   USE prtctl          ! Print control


   IMPLICIT NONE
   PRIVATE

   PUBLIC   tra_adv_tvd    ! routine called by step.F90

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !!   OPA 9.0 , LOCEAN-IPSL (2006) 
   !! $Id$
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS

   SUBROUTINE tra_adv_tvd( kt, pun, pvn, pwn )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE tra_adv_tvd  ***
      !! 
      !! **  Purpose :   Compute the now trend due to total advection of 
      !!       tracers and add it to the general trend of tracer equations
      !!
      !! **  Method  :   TVD scheme, i.e. 2nd order centered scheme with
      !!       corrected flux (monotonic correction)
      !!       note: - this advection scheme needs a leap-frog time scheme
      !!
      !! ** Action : - update (ta,sa) with the now advective tracer trends
      !!             - save the trends in (ztrdt,ztrds) ('key_trdtra')
      !!----------------------------------------------------------------------
      USE oce              , ztrdt => ua   ! use ua as workspace
      USE oce              , ztrds => va   ! use va as workspace
      !!
      INTEGER , INTENT(in)                         ::   kt    ! ocean time-step index
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pun   ! ocean velocity u-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pvn   ! ocean velocity v-component
      REAL(wp), INTENT(in), DIMENSION(jpi,jpj,jpk) ::   pwn   ! ocean velocity w-component
      !!
      INTEGER  ::   ji, jj, jk              ! dummy loop indices
      REAL(wp) ::                        &  ! temporary scalar
         ztat, zsat,                     &  !    "         "   
         z_hdivn_x, z_hdivn_y, z_hdivn
      REAL(wp) ::   &
         z2dtt, zbtr, zeu, zev,          &  ! temporary scalar
         zew, z2, zbtr1,                 &  ! temporary scalar
         zfp_ui, zfp_vj, zfp_wk,         &  !    "         "
         zfm_ui, zfm_vj, zfm_wk             !    "         "
      REAL(wp), DIMENSION (jpi,jpj,jpk) ::   zti, ztu, ztv, ztw   ! temporary workspace
      REAL(wp), DIMENSION (jpi,jpj,jpk) ::   zsi, zsu, zsv, zsw   !    "           "
      !!----------------------------------------------------------------------

      zti(:,:,:) = 0.e0   ;   zsi(:,:,:) = 0.e0

      IF( kt == nit000 .AND. lwp ) THEN
         WRITE(numout,*)
         WRITE(numout,*) 'tra_adv_tvd : TVD advection scheme'
         WRITE(numout,*) '~~~~~~~~~~~'
      ENDIF

      IF( neuler == 0 .AND. kt == nit000 ) THEN   ;    z2 = 1.
      ELSE                                        ;    z2 = 2.
      ENDIF

      ! 1. Bottom value : flux set to zero
      ! ---------------
      ztu(:,:,jpk) = 0.e0   ;   zsu(:,:,jpk) = 0.e0
      ztv(:,:,jpk) = 0.e0   ;   zsv(:,:,jpk) = 0.e0
      ztw(:,:,jpk) = 0.e0   ;   zsw(:,:,jpk) = 0.e0
      zti(:,:,jpk) = 0.e0   ;   zsi(:,:,jpk) = 0.e0


      ! 2. upstream advection with initial mass fluxes & intermediate update
      ! --------------------------------------------------------------------
      ! upstream tracer flux in the i and j direction
      DO jk = 1, jpkm1
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zeu = 0.5 * e2u(ji,jj) * fse3u(ji,jj,jk) * pun(ji,jj,jk)
               zev = 0.5 * e1v(ji,jj) * fse3v(ji,jj,jk) * pvn(ji,jj,jk)
               ! upstream scheme
               zfp_ui = zeu + ABS( zeu )
               zfm_ui = zeu - ABS( zeu )
               zfp_vj = zev + ABS( zev )
               zfm_vj = zev - ABS( zev )
               ztu(ji,jj,jk) = zfp_ui * tb(ji,jj,jk) + zfm_ui * tb(ji+1,jj  ,jk)
               ztv(ji,jj,jk) = zfp_vj * tb(ji,jj,jk) + zfm_vj * tb(ji  ,jj+1,jk)
               zsu(ji,jj,jk) = zfp_ui * sb(ji,jj,jk) + zfm_ui * sb(ji+1,jj  ,jk)
               zsv(ji,jj,jk) = zfp_vj * sb(ji,jj,jk) + zfm_vj * sb(ji  ,jj+1,jk)
            END DO
         END DO
      END DO

      ! upstream tracer flux in the k direction
      ! Surface value
      IF( lk_vvl ) THEN
         ! variable volume : flux set to zero
         ztw(:,:,1) = 0.e0
         zsw(:,:,1) = 0.e0
      ELSE
         ! free surface-constant volume
         DO jj = 1, jpj
            DO ji = 1, jpi
               zew = e1t(ji,jj) * e2t(ji,jj) * pwn(ji,jj,1)
               ztw(ji,jj,1) = zew * tb(ji,jj,1)
               zsw(ji,jj,1) = zew * sb(ji,jj,1)
            END DO
         END DO
      ENDIF

      ! Interior value
      DO jk = 2, jpkm1
         DO jj = 1, jpj
            DO ji = 1, jpi
               zew = 0.5 * e1t(ji,jj) * e2t(ji,jj) * pwn(ji,jj,jk)
               zfp_wk = zew + ABS( zew )
               zfm_wk = zew - ABS( zew )
               ztw(ji,jj,jk) = zfp_wk * tb(ji,jj,jk) + zfm_wk * tb(ji,jj,jk-1)
               zsw(ji,jj,jk) = zfp_wk * sb(ji,jj,jk) + zfm_wk * sb(ji,jj,jk-1)
            END DO
         END DO
      END DO

      ! total advective trend
      DO jk = 1, jpkm1
         z2dtt = z2 * rdttra(jk)
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zbtr = 1./ ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
               ! total intermediate advective trends
               ztat = - ( ztu(ji,jj,jk) - ztu(ji-1,jj  ,jk  )   &
                  &     + ztv(ji,jj,jk) - ztv(ji  ,jj-1,jk  )   &
                  &     + ztw(ji,jj,jk) - ztw(ji  ,jj  ,jk+1) ) * zbtr
               zsat = - ( zsu(ji,jj,jk) - zsu(ji-1,jj  ,jk  )   & 
                  &     + zsv(ji,jj,jk) - zsv(ji  ,jj-1,jk  )   &
                  &     + zsw(ji,jj,jk) - zsw(ji  ,jj  ,jk+1) ) * zbtr
               ! update and guess with monotonic sheme
               ta(ji,jj,jk) =  ta(ji,jj,jk) + ztat
               sa(ji,jj,jk) =  sa(ji,jj,jk) + zsat
               zti (ji,jj,jk) = ( tb(ji,jj,jk) + z2dtt * ztat ) * tmask(ji,jj,jk)
               zsi (ji,jj,jk) = ( sb(ji,jj,jk) + z2dtt * zsat ) * tmask(ji,jj,jk)
            END DO
         END DO
      END DO

      ! "zonal" mean advective heat and salt transport
      IF( ln_diaptr .AND. ( MOD( kt, nf_ptr ) == 0 ) ) THEN
         pht_adv(:) = ptr_vj( ztv(:,:,:) )
         pst_adv(:) = ptr_vj( zsv(:,:,:) )
      ENDIF

      ! Save the intermediate i / j / k advective trends for diagnostics
      ! -------------------------------------------------------------------
      ! Warning : We should use zun instead of un in the computations below, but we
      ! also use hdivn which is computed with un, vn (check ???). So we use un, vn
      ! for consistency. Results are therefore approximate with key_trabbl_adv.

      IF( l_trdtra ) THEN
         ztrdt(:,:,:) = 0.e0   ;   ztrds(:,:,:) = 0.e0
         ! 
         ! T/S ZONAL advection trends
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zbtr = 1. / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
                  ztrdt(ji,jj,jk) = - ( ztu(ji,jj,jk) - ztu(ji-1,jj,jk) ) * zbtr
                  ztrds(ji,jj,jk) = - ( zsu(ji,jj,jk) - zsu(ji-1,jj,jk) ) * zbtr
               END DO
            END DO
         END DO
         CALL trd_mod(ztrdt, ztrds, jptra_trd_xad, 'TRA', kt)    ! save the trends
         !
         ! T/S MERIDIONAL advection trends
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zbtr      = 1. / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
                  ztrdt(ji,jj,jk) = - ( ztv(ji,jj,jk) - ztv(ji,jj-1,jk) ) * zbtr
                  ztrds(ji,jj,jk) = - ( zsv(ji,jj,jk) - zsv(ji,jj-1,jk) ) * zbtr
               END DO
            END DO
         END DO
         CALL trd_mod(ztrdt, ztrds, jptra_trd_yad, 'TRA', kt)     ! save the trends
         !
         ! T/S VERTICAL advection trends
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.         
                  zbtr      = 1. / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
                  ztrdt(ji,jj,jk) = - ( ztw(ji,jj,jk) - ztw(ji,jj,jk+1) ) * zbtr
                  ztrds(ji,jj,jk) = - ( zsw(ji,jj,jk) - zsw(ji,jj,jk+1) ) * zbtr
               END DO
            END DO
         END DO
         CALL trd_mod(ztrdt, ztrds, jptra_trd_zad, 'TRA', kt)     ! save the trends
         !
      ENDIF

      ! Lateral boundary conditions on zti, zsi   (unchanged sign)
      CALL lbc_lnk( zti, 'T', 1. )
      CALL lbc_lnk( zsi, 'T', 1. )


      ! 3. antidiffusive flux : high order minus low order
      ! --------------------------------------------------
      ! antidiffusive flux on i and j
      DO jk = 1, jpkm1
         DO jj = 1, jpjm1
            DO ji = 1, fs_jpim1   ! vector opt.
               zeu = 0.5 * e2u(ji,jj) * fse3u(ji,jj,jk) * pun(ji,jj,jk)
               zev = 0.5 * e1v(ji,jj) * fse3v(ji,jj,jk) * pvn(ji,jj,jk)
               ztu(ji,jj,jk) = zeu * ( tn(ji,jj,jk) + tn(ji+1,jj,jk) ) - ztu(ji,jj,jk)
               zsu(ji,jj,jk) = zeu * ( sn(ji,jj,jk) + sn(ji+1,jj,jk) ) - zsu(ji,jj,jk)
               ztv(ji,jj,jk) = zev * ( tn(ji,jj,jk) + tn(ji,jj+1,jk) ) - ztv(ji,jj,jk)
               zsv(ji,jj,jk) = zev * ( sn(ji,jj,jk) + sn(ji,jj+1,jk) ) - zsv(ji,jj,jk)
            END DO
         END DO
      END DO
      
      ! antidiffusive flux on k
      ! Surface value
      ztw(:,:,1) = 0.e0
      zsw(:,:,1) = 0.e0

      ! Interior value
      DO jk = 2, jpkm1
         DO jj = 1, jpj
            DO ji = 1, jpi
               zew = 0.5 * e1t(ji,jj) * e2t(ji,jj) * pwn(ji,jj,jk)
               ztw(ji,jj,jk) = zew * ( tn(ji,jj,jk) + tn(ji,jj,jk-1) ) - ztw(ji,jj,jk)
               zsw(ji,jj,jk) = zew * ( sn(ji,jj,jk) + sn(ji,jj,jk-1) ) - zsw(ji,jj,jk)
            END DO
         END DO
      END DO

      ! Lateral bondary conditions
      CALL lbc_lnk( ztu, 'U', -1. )   ;   CALL lbc_lnk( zsu, 'U', -1. )
      CALL lbc_lnk( ztv, 'V', -1. )   ;   CALL lbc_lnk( zsv, 'V', -1. )
      CALL lbc_lnk( ztw, 'W',  1. )   ;   CALL lbc_lnk( zsw, 'W',  1. )

      ! 4. monotonicity algorithm
      ! -------------------------
      CALL nonosc( tb, ztu, ztv, ztw, zti, z2 )
      CALL nonosc( sb, zsu, zsv, zsw, zsi, z2 )


      ! 5. final trend with corrected fluxes
      ! ------------------------------------
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.  
               zbtr = 1. / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
               ! total advective trends
               ztat = - ( ztu(ji,jj,jk) - ztu(ji-1,jj  ,jk  )   &
                  &     + ztv(ji,jj,jk) - ztv(ji  ,jj-1,jk  )   &
                  &     + ztw(ji,jj,jk) - ztw(ji  ,jj  ,jk+1) ) * zbtr
               zsat = - ( zsu(ji,jj,jk) - zsu(ji-1,jj  ,jk  )   &
                  &     + zsv(ji,jj,jk) - zsv(ji  ,jj-1,jk  )   &
                  &     + zsw(ji,jj,jk) - zsw(ji  ,jj  ,jk+1) ) * zbtr
               ! add them to the general tracer trends
               ta(ji,jj,jk) = ta(ji,jj,jk) + ztat
               sa(ji,jj,jk) = sa(ji,jj,jk) + zsat
            END DO
         END DO
      END DO


      ! Save the advective trends for diagnostics
      ! --------------------------------------------

      IF( l_trdtra ) THEN
         ztrdt(:,:,:) = 0.e0   ;   ztrds(:,:,:) = 0.e0
         !
         ! T/S ZONAL advection trends
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  !-- Compute zonal divergence by splitting hdivn (see divcur.F90)
                  !   N.B. This computation is not valid along OBCs (if any)
                  zbtr = 1. / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
                  z_hdivn_x = (  e2u(ji  ,jj) * fse3u(ji  ,jj,jk) * pun(ji  ,jj,jk)          &
                     &         - e2u(ji-1,jj) * fse3u(ji-1,jj,jk) * pun(ji-1,jj,jk) ) * zbtr
                  !-- Compute T/S zonal advection trends
                  ztrdt(ji,jj,jk) = - ( ztu(ji,jj,jk) - ztu(ji-1,jj,jk) ) * zbtr + tn(ji,jj,jk) * z_hdivn_x
                  ztrds(ji,jj,jk) = - ( zsu(ji,jj,jk) - zsu(ji-1,jj,jk) ) * zbtr + sn(ji,jj,jk) * z_hdivn_x
               END DO
            END DO
         END DO
         CALL trd_mod(ztrdt, ztrds, jptra_trd_xad, 'TRA', kt, cnbpas='bis')   ! <<< ADD TO PREVIOUSLY COMPUTED
         !
         ! T/S MERIDIONAL advection trends
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  !-- Compute merid. divergence by splitting hdivn (see divcur.F90)
                  !   N.B. This computation is not valid along OBCs (if any)
                  zbtr      = 1. / ( e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) )
                  z_hdivn_y = (  e1v(ji,  jj) * fse3v(ji,jj  ,jk) * pvn(ji,jj  ,jk)          &
                     &         - e1v(ji,jj-1) * fse3v(ji,jj-1,jk) * pvn(ji,jj-1,jk) ) * zbtr
                  !-- Compute T/S meridional advection trends
                  ztrdt(ji,jj,jk) = - ( ztv(ji,jj,jk) - ztv(ji,jj-1,jk) ) * zbtr + tn(ji,jj,jk) * z_hdivn_y          
                  ztrds(ji,jj,jk) = - ( zsv(ji,jj,jk) - zsv(ji,jj-1,jk) ) * zbtr + sn(ji,jj,jk) * z_hdivn_y          
               END DO
            END DO
         END DO
         CALL trd_mod(ztrdt, ztrds, jptra_trd_yad, 'TRA', kt, cnbpas='bis')   ! <<< ADD TO PREVIOUSLY COMPUTED
         !
         ! T/S VERTICAL advection trends
         DO jk = 1, jpkm1
            DO jj = 2, jpjm1
               DO ji = fs_2, fs_jpim1   ! vector opt.
                  zbtr1     = 1. / ( e1t(ji,jj) * e2t(ji,jj) )
#if defined key_zco
                  zbtr      = zbtr1
                  z_hdivn_x = e2u(ji,jj)*pun(ji,jj,jk) - e2u(ji-1,jj)*pun(ji-1,jj,jk)
                  z_hdivn_y = e1v(ji,jj)*pvn(ji,jj,jk) - e1v(ji,jj-1)*pvn(ji,jj-1,jk)
#else
                  zbtr      = zbtr1 / fse3t(ji,jj,jk)
                  z_hdivn_x = e2u(ji,jj)*fse3u(ji,jj,jk)*pun(ji,jj,jk) - e2u(ji-1,jj)*fse3u(ji-1,jj,jk)*pun(ji-1,jj,jk)
                  z_hdivn_y = e1v(ji,jj)*fse3v(ji,jj,jk)*pvn(ji,jj,jk) - e1v(ji,jj-1)*fse3v(ji,jj-1,jk)*pvn(ji,jj-1,jk)
#endif
                  z_hdivn   = (z_hdivn_x + z_hdivn_y) * zbtr
                  zbtr      = zbtr1 / fse3t(ji,jj,jk)
                  ztrdt(ji,jj,jk) = - ( ztw(ji,jj,jk) - ztw(ji,jj,jk+1) ) * zbtr - tn(ji,jj,jk) * z_hdivn
                  ztrds(ji,jj,jk) = - ( zsw(ji,jj,jk) - zsw(ji,jj,jk+1) ) * zbtr - sn(ji,jj,jk) * z_hdivn
               END DO
            END DO
         END DO
         CALL trd_mod(ztrdt, ztrds, jptra_trd_zad, 'TRA', kt, cnbpas='bis')   ! <<< ADD TO PREVIOUSLY COMPUTED
         !
      ENDIF

      IF(ln_ctl)   CALL prt_ctl( tab3d_1=ta, clinfo1=' tvd adv  - Ta: ', mask1=tmask,   &
         &                       tab3d_2=sa, clinfo2=           ' Sa: ', mask2=tmask, clinfo3='tra' )

      ! "zonal" mean advective heat and salt transport
      IF( ln_diaptr .AND. ( MOD( kt, nf_ptr ) == 0 ) ) THEN
         pht_adv(:) = ptr_vj( ztv(:,:,:) ) + pht_adv(:)
         pst_adv(:) = ptr_vj( zsv(:,:,:) ) + pst_adv(:)
      ENDIF
      !
   END SUBROUTINE tra_adv_tvd


   SUBROUTINE nonosc( pbef, paa, pbb, pcc, paft, prdt )
      !!---------------------------------------------------------------------
      !!                    ***  ROUTINE nonosc  ***
      !!     
      !! **  Purpose :   compute monotonic tracer fluxes from the upstream 
      !!       scheme and the before field by a nonoscillatory algorithm 
      !!
      !! **  Method  :   ... ???
      !!       warning : pbef and paft must be masked, but the boundaries
      !!       conditions on the fluxes are not necessary zalezak (1979)
      !!       drange (1995) multi-dimensional forward-in-time and upstream-
      !!       in-space based differencing for fluid
      !!----------------------------------------------------------------------
      REAL(wp), INTENT( in ) ::   prdt   ! ???
      REAL(wp), DIMENSION (jpi,jpj,jpk), INTENT( inout ) ::   &
         pbef,                            & ! before field
         paft,                            & ! after field
         paa,                             & ! monotonic flux in the i direction
         pbb,                             & ! monotonic flux in the j direction
         pcc                                ! monotonic flux in the k direction
      !!
      INTEGER ::   ji, jj, jk               ! dummy loop indices
      INTEGER ::   ikm1
      REAL(wp), DIMENSION (jpi,jpj,jpk) ::   zbetup, zbetdo
      REAL(wp), DIMENSION (jpi,jpj,jpk) ::   zbup, zbdo
      REAL(wp) ::   zpos, zneg, zbt, za, zb, zc, zbig, zrtrn, z2dtt
      REAL(wp) ::   zau, zbu, zcu, zav, zbv, zcv
      REAL(wp) ::   zup, zdo
      !!----------------------------------------------------------------------

      zbig = 1.e+40
      zrtrn = 1.e-15
      zbetup(:,:,jpk) = 0.e0   ;   zbetdo(:,:,jpk) = 0.e0


      ! Search local extrema
      ! --------------------
      ! max/min of pbef & paft with large negative/positive value (-/+zbig) inside land
      zbup = MAX( pbef * tmask - zbig * ( 1.e0 - tmask ),   &
         &        paft * tmask - zbig * ( 1.e0 - tmask )  )
      zbdo = MIN( pbef * tmask + zbig * ( 1.e0 - tmask ),   &
         &        paft * tmask + zbig * ( 1.e0 - tmask )  )

      DO jk = 1, jpkm1
         ikm1 = MAX(jk-1,1)
         z2dtt = prdt * rdttra(jk)
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.

               ! search maximum in neighbourhood
               zup = MAX(  zbup(ji  ,jj  ,jk  ),   &
                  &        zbup(ji-1,jj  ,jk  ), zbup(ji+1,jj  ,jk  ),   &
                  &        zbup(ji  ,jj-1,jk  ), zbup(ji  ,jj+1,jk  ),   &
                  &        zbup(ji  ,jj  ,ikm1), zbup(ji  ,jj  ,jk+1)  )

               ! search minimum in neighbourhood
               zdo = MIN(  zbdo(ji  ,jj  ,jk  ),   &
                  &        zbdo(ji-1,jj  ,jk  ), zbdo(ji+1,jj  ,jk  ),   &
                  &        zbdo(ji  ,jj-1,jk  ), zbdo(ji  ,jj+1,jk  ),   &
                  &        zbdo(ji  ,jj  ,ikm1), zbdo(ji  ,jj  ,jk+1)  )

               ! positive part of the flux
               zpos = MAX( 0., paa(ji-1,jj  ,jk  ) ) - MIN( 0., paa(ji  ,jj  ,jk  ) )   &
                  & + MAX( 0., pbb(ji  ,jj-1,jk  ) ) - MIN( 0., pbb(ji  ,jj  ,jk  ) )   &
                  & + MAX( 0., pcc(ji  ,jj  ,jk+1) ) - MIN( 0., pcc(ji  ,jj  ,jk  ) )

               ! negative part of the flux
               zneg = MAX( 0., paa(ji  ,jj  ,jk  ) ) - MIN( 0., paa(ji-1,jj  ,jk  ) )   &
                  & + MAX( 0., pbb(ji  ,jj  ,jk  ) ) - MIN( 0., pbb(ji  ,jj-1,jk  ) )   &
                  & + MAX( 0., pcc(ji  ,jj  ,jk  ) ) - MIN( 0., pcc(ji  ,jj  ,jk+1) )

               ! up & down beta terms
               zbt = e1t(ji,jj) * e2t(ji,jj) * fse3t(ji,jj,jk) / z2dtt
               zbetup(ji,jj,jk) = ( zup            - paft(ji,jj,jk) ) / ( zpos + zrtrn ) * zbt
               zbetdo(ji,jj,jk) = ( paft(ji,jj,jk) - zdo            ) / ( zneg + zrtrn ) * zbt
            END DO
         END DO
      END DO

      ! lateral boundary condition on zbetup & zbetdo   (unchanged sign)
      CALL lbc_lnk( zbetup, 'T', 1. )
      CALL lbc_lnk( zbetdo, 'T', 1. )


      ! 3. monotonic flux in the i & j direction (paa & pbb)
      ! ----------------------------------------
      DO jk = 1, jpkm1
         DO jj = 2, jpjm1
            DO ji = fs_2, fs_jpim1   ! vector opt.
               zau = MIN( 1.e0, zbetdo(ji,jj,jk), zbetup(ji+1,jj,jk) )
               zbu = MIN( 1.e0, zbetup(ji,jj,jk), zbetdo(ji+1,jj,jk) )
               zcu =       ( 0.5  + SIGN( 0.5 , paa(ji,jj,jk) ) )
               paa(ji,jj,jk) = paa(ji,jj,jk) * ( zcu * zau + ( 1.e0 - zcu) * zbu )

               zav = MIN( 1.e0, zbetdo(ji,jj,jk), zbetup(ji,jj+1,jk) )
               zbv = MIN( 1.e0, zbetup(ji,jj,jk), zbetdo(ji,jj+1,jk) )
               zcv =       ( 0.5  + SIGN( 0.5 , pbb(ji,jj,jk) ) )
               pbb(ji,jj,jk) = pbb(ji,jj,jk) * ( zcv * zav + ( 1.e0 - zcv) * zbv )

      ! monotonic flux in the k direction, i.e. pcc
      ! -------------------------------------------
               za = MIN( 1., zbetdo(ji,jj,jk+1), zbetup(ji,jj,jk) )
               zb = MIN( 1., zbetup(ji,jj,jk+1), zbetdo(ji,jj,jk) )
               zc =       ( 0.5  + SIGN( 0.5 , pcc(ji,jj,jk+1) ) )
               pcc(ji,jj,jk+1) = pcc(ji,jj,jk+1) * ( zc * za + ( 1.e0 - zc) * zb )
            END DO
         END DO
      END DO

      ! lateral boundary condition on paa, pbb, pcc
      CALL lbc_lnk( paa, 'U', -1. )      ! changed sign
      CALL lbc_lnk( pbb, 'V', -1. )      ! changed sign
      !
   END SUBROUTINE nonosc

   !!======================================================================
END MODULE traadv_tvd
